# import datetime
import jinja2
import os
import shutil
import time
from ..buildjson import buildjson
from ..copyright import copyright_manager
import json2daisy
from . import parameters


class c2daisy:
    """ Generates a Daisy wrapper for a given patch.
    """

    @classmethod
    def compile(clazz, c_src_dir, out_dir, externs, patch_name=None, patch_meta: dict = None,
                num_input_channels=0, num_output_channels=0, copyright=None, verbose=False):

        tick = time.time()

        out_dir = os.path.join(out_dir, "daisy")

        if patch_meta:
            patch_name = patch_meta.get("name", patch_name)
            daisy_meta = patch_meta.get("daisy")
        else:
            daisy_meta = {}

        board = daisy_meta.get("board", "seed")

        copyright_c = copyright_manager.get_copyright_for_c(copyright)
        # copyright_plist = copyright or u"Copyright {0} Enzien Audio, Ltd." \
        #     " All Rights Reserved.".format(datetime.datetime.now().year)

        try:
            # ensure that the output directory does not exist
            out_dir = os.path.abspath(out_dir)
            if os.path.exists(out_dir):
                shutil.rmtree(out_dir)

            # copy over static files
            shutil.copytree(os.path.join(os.path.dirname(__file__), "static"), out_dir)

            # copy over generated C source files
            source_dir = os.path.join(out_dir, "source")
            shutil.copytree(c_src_dir, source_dir)

            if daisy_meta.get('board_file'):
                header, board_info = json2daisy.generate_header_from_file(daisy_meta['board_file'])
            else:
                header, board_info = json2daisy.generate_header_from_name(board)

            component_glue = parameters.parse_parameters(
                externs['parameters'], board_info['components'], board_info['aliases'], 'hardware')
            component_glue['class_name'] = board_info['name']
            component_glue['patch_name'] = patch_name
            component_glue['header'] = f"HeavyDaisy_{patch_name}.hpp"
            component_glue['max_channels'] = board_info['channels']
            component_glue['num_output_channels'] = num_output_channels

            component_glue['copyright'] = copyright_c

            daisy_h_path = os.path.join(source_dir, f"HeavyDaisy_{patch_name}.hpp")
            with open(daisy_h_path, "w") as f:
                f.write(header)

            loader = jinja2.FileSystemLoader(os.path.join(os.path.dirname(os.path.abspath(__file__)), 'templates'))
            env = jinja2.Environment(loader=loader, trim_blocks=True, lstrip_blocks=True)
            daisy_cpp_path = os.path.join(source_dir, f"HeavyDaisy_{patch_name}.cpp")

            rendered_cpp = env.get_template('HeavyDaisy.cpp').render(component_glue)
            with open(daisy_cpp_path, 'w') as file:
                file.write(rendered_cpp)

            makefile_replacements = {'name': patch_name}
            makefile_replacements['linker_script'] = daisy_meta.get('linker_script', '')
            if makefile_replacements['linker_script'] != '':
                makefile_replacements['linker_script'] = f'../{daisy_meta["linker_script"]}'
            depth = daisy_meta.get('libdaisy_depth', 2)
            makefile_replacements['libdaisy_path'] = f'{"../" * depth}libdaisy'
            makefile_replacements['bootloader'] = daisy_meta.get('bootloader', False)

            rendered_makefile = env.get_template('Makefile').render(makefile_replacements)
            with open(os.path.join(source_dir, "Makefile"), "w") as f:
                f.write(rendered_makefile)

            # ======================================================================================

            buildjson.generate_json(
                out_dir,
                linux_x64_args=["-j"])

            return {
                "stage": "c2daisy",
                "notifs": {
                    "has_error": False,
                    "exception": None,
                    "warnings": [],
                    "errors": []
                },
                "in_dir": c_src_dir,
                "in_file": "",
                "out_dir": out_dir,
                "out_file": os.path.basename(daisy_h_path),
                "compile_time": time.time() - tick
            }

        except Exception as e:
            return {
                "stage": "c2daisy",
                "notifs": {
                    "has_error": True,
                    "exception": e,
                    "warnings": [],
                    "errors": [{
                        "enum": -1,
                        "message": str(e)
                    }]
                },
                "in_dir": c_src_dir,
                "in_file": "",
                "out_dir": out_dir,
                "out_file": "",
                "compile_time": time.time() - tick
            }



# import datetime
import os
import shutil
import time
import jinja2
import json
from ..buildjson import buildjson
from ..copyright import copyright_manager
import hvcc.core.hv2ir.HeavyLangObject as HeavyLangObject


heavy_hash = HeavyLangObject.HeavyLangObject.get_hash
OWL_BUTTONS = ['Push', 'B1', 'B2', 'B3', 'B4', 'B5', 'B6', 'B7', 'B8']


class c2owl:
    """ Generates a OWL wrapper for a given patch.
    """

    @classmethod
    def make_jdata(clazz, patch_ir):
        jdata = list()

        with open(patch_ir, mode="r") as f:
            ir = json.load(f)

            for name, v in ir['control']['receivers'].items():
                # skip __hv_init and similar
                if name.startswith("__"):
                    continue

                # If a name has been specified
                if v['attributes'].get('raw'):
                    key = v['attributes']['raw']
                    jdata.append((key, name, 'RECV', f"0x{heavy_hash(name):X}",
                                  v['attributes']['min'],
                                  v['attributes']['max'],
                                  v['attributes']['default'],
                                  key in OWL_BUTTONS))

                elif name.startswith('Channel-'):
                    key = name.split('Channel-', 1)[1]
                    jdata.append((key, name, 'RECV', f"0x{heavy_hash(name):X}",
                                  0, 1, None, key in OWL_BUTTONS))

            for k, v in ir['objects'].items():
                try:
                    if v['type'] == '__send':
                        name = v['args']['name']
                        if v['args']['attributes'].get('raw'):
                            key = v['args']['attributes']['raw']
                            jdata.append((key, f'{name}>', 'SEND', f"0x{heavy_hash(name):X}",
                                          v['args']['attributes']['min'],
                                          v['args']['attributes']['max'],
                                          v['args']['attributes']['default'],
                                          key in OWL_BUTTONS))
                        elif name.startswith('Channel-'):
                            key = name.split('Channel-', 1)[1]
                            jdata.append((key, f'{name}>', 'SEND', f"0x{heavy_hash(name):X}",
                                          0, 1, None, key in OWL_BUTTONS))
                except Exception:
                    pass

            return jdata

    @classmethod
    def compile(clazz, c_src_dir, out_dir, externs, patch_name=None, patch_meta: dict = None,
                num_input_channels=0, num_output_channels=0, copyright=None, verbose=False):

        tick = time.time()

        out_dir = os.path.join(out_dir, "Source")
        patch_name = patch_name or "heavy"
        copyright_c = copyright_manager.get_copyright_for_c(copyright)

        try:
            # ensure that the output directory does not exist
            out_dir = os.path.abspath(out_dir)
            if os.path.exists(out_dir):
                shutil.rmtree(out_dir)

            # copy over generated C source files
            shutil.copytree(c_src_dir, out_dir)

            # copy over deps
            shutil.copytree(os.path.join(os.path.dirname(__file__), "deps"), out_dir, dirs_exist_ok=True)

            # initialize the jinja template environment
            env = jinja2.Environment()

            env.loader = jinja2.FileSystemLoader(
                os.path.join(os.path.dirname(os.path.abspath(__file__)), "templates"))

            # construct jdata from ir
            ir_dir = os.path.join(c_src_dir, "../ir")
            patch_ir = os.path.join(ir_dir, f"{patch_name}.heavy.ir.json")
            jdata = c2owl.make_jdata(patch_ir)

            # generate OWL wrapper from template
            owl_hpp_path = os.path.join(out_dir, f"HeavyOWL_{patch_name}.hpp")
            with open(owl_hpp_path, "w") as f:
                f.write(env.get_template("HeavyOwl.hpp").render(
                    jdata=jdata,
                    name=patch_name,
                    copyright=copyright_c))
            owl_h_path = os.path.join(out_dir, "HeavyOwlConstants.h")
            with open(owl_h_path, "w") as f:
                f.write(env.get_template("HeavyOwlConstants.h").render(
                    jdata=jdata,
                    copyright=copyright_c))

            # generate list of Heavy source files
            # files = os.listdir(out_dir)

            # ======================================================================================
            # Linux
            #
            # linux_path = os.path.join(out_dir, "linux")
            # os.makedirs(linux_path)

            # with open(os.path.join(out_dir, "Makefile"), "w") as f:
            #     f.write(env.get_template("Makefile").render(
            #         name=patch_name,
            #         class_name=f"HeavyOWL_{patch_name}"))

            buildjson.generate_json(
                out_dir,
                linux_x64_args=["-j"])
            # macos_x64_args=["-project", "{0}.xcodeproj".format(patch_name), "-arch",
            #                 "x86_64", "-alltargets"],
            # win_x64_args=["/property:Configuration=Release", "/property:Platform=x64",
            #               "/t:Rebuild", "{0}.sln".format(patch_name), "/m"],
            # win_x86_args=["/property:Configuration=Release", "/property:Platform=x86",
            #               "/t:Rebuild", "{0}.sln".format(patch_name), "/m"])

            return {
                "stage": "c2owl",
                "notifs": {
                    "has_error": False,
                    "exception": None,
                    "warnings": [],
                    "errors": []
                },
                "in_dir": c_src_dir,
                "in_file": "",
                "out_dir": out_dir,
                "out_file": os.path.basename(owl_h_path),
                "compile_time": time.time() - tick
            }

        except Exception as e:
            return {
                "stage": "c2owl",
                "notifs": {
                    "has_error": True,
                    "exception": e,
                    "warnings": [],
                    "errors": [{
                        "enum": -1,
                        "message": str(e)
                    }]
                },
                "in_dir": c_src_dir,
                "in_file": "",
                "out_dir": out_dir,
                "out_file": "",
                "compile_time": time.time() - tick
            }



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .HeavyObject import HeavyObject


class ControlPack(HeavyObject):
    """Prints the first value in a message to the console"""

    c_struct = "ControlPack"
    preamble = "cPack"

    @classmethod
    def get_C_header_set(clazz):
        return {"HvControlPack.h"}

    @classmethod
    def get_C_file_set(clazz):
        return {"HvControlPack.h", "HvControlPack.c"}

    @classmethod
    def get_C_init(clazz, obj_type, obj_id, args):
        return [
            "cPack_init(&cPack_{0}, {1}, {2});".format(
                obj_id,
                len(args["values"]),
                ", ".join(["{0}f".format(float(x)) for x in args["values"]]))
        ]

    @classmethod
    def get_C_onMessage(clazz, obj_type, obj_id, inlet_index, args):
        return [
            "cPack_onMessage(_c, &Context(_c)->cPack_{0}, {1}, m, &cPack_{0}_sendMessage);".format(
                obj_id,
                inlet_index)
        ]



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .HeavyObject import HeavyObject


class SignalTabread(HeavyObject):
    """Handles __tabread~if, __tabread~f, __tabreadu~f
    """

    c_struct = "SignalTabread"
    preamble = "sTabread"

    @classmethod
    def get_C_header_set(clazz):
        return {"HvSignalTabread.h"}

    @classmethod
    def get_C_file_set(clazz):
        return {"HvSignalTabread.h", "HvSignalTabread.c"}

    @classmethod
    def get_C_free(clazz, obj_type, obj_id, args):
        return []

    @classmethod
    def get_C_init(clazz, obj_type, obj_id, args):
        return [
            "sTabread_init(&sTabread_{0}, &hTable_{1}, {2});".format(
                obj_id,
                args["table_id"],
                "true" if obj_type == "__tabread~f" else "false")]

    @classmethod
    def get_C_onMessage(clazz, obj_type, obj_id, inlet_index, args):
        if obj_type in ["__tabread~f", "__tabreadu~f"]:
            return [
                "sTabread_onMessage(_c, &Context(_c)->sTabread_{0}, {1}, m, &sTabread_{0}_sendMessage);".format(
                    obj_id,
                    inlet_index)]
        else:  # "__tabread~if"
            return [
                "sTabread_onMessage(_c, &Context(_c)->sTabread_{0}, {1}, m, NULL);".format(
                    obj_id,
                    inlet_index)]

    @classmethod
    def get_C_process(clazz, process_dict, obj_type, obj_id, args):
        if obj_type == "__tabread~if":
            return [
                "__hv_tabread_if(&sTabread_{0}, {1}, {2});".format(
                    process_dict["id"],
                    ", ".join(["VIi({0})".format(HeavyObject._c_buffer(b)) for b in process_dict["inputBuffers"]]),
                    ", ".join(["VOf({0})".format(HeavyObject._c_buffer(b)) for b in process_dict["outputBuffers"]])
                )]
        elif obj_type == "__tabread~f":
            return [
                "__hv_tabread_f(&sTabread_{0}, {1});".format(
                    process_dict["id"],
                    ", ".join(["VOf({0})".format(HeavyObject._c_buffer(b)) for b in process_dict["outputBuffers"]])
                )]
        elif obj_type == "__tabreadu~f":
            return [
                "__hv_tabreadu_f(&sTabread_{0}, {1});".format(
                    process_dict["id"],
                    ", ".join(["VOf({0})".format(HeavyObject._c_buffer(b)) for b in process_dict["outputBuffers"]])
                )]
        else:
            raise Exception()
